#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-08-16
# @Filename: time.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import ctypes
import datetime
import time

import numpy

from . import sofa
from .exceptions import CoordIOError
from .iers import IERS


def utc_to_tai(jd):
    """Convert UTC to TAI. Wrapper around iauUtctai."""

    utc1 = int(jd)
    utc2 = jd - utc1

    tai1 = ctypes.c_double()
    tai2 = ctypes.c_double()

    res = sofa.iauUtctai(utc1, utc2, tai1, tai2)

    if res != 0:
        raise ValueError(f'iauUtctai return with error code {res}.')

    return tai1.value + tai2.value


class Time:
    """A time storage and conversion class.

    All times, regardless of the input format and scale are converted and
    stored as TAI Julian dates.

    Parameters
    ----------
    value
        The input time value. If `None`, the current `datetime.datetime.now`
        system time will be used.
    format : str
        The format of the input time.
    scale : str
        The time scale of the input value. Valid values are ``UTC`` and
        ``TAI``.

    Attributes
    ----------
    jd : float
        The Julian date in the TAI scale.

    """

    def __init__(self, value=None, format='jd', scale='UTC'):

        if isinstance(value, datetime.datetime) or value is None:
            jd1, jd2 = sofa.get_internal_date(value, scale)
            self.jd = jd1 + jd2
        elif format == 'jd':
            self.jd = value
        else:
            raise NotImplementedError('Not implemented format.')

        if scale == 'UTC':
            self.jd = utc_to_tai(self.jd)
        elif scale == 'TAI':
            pass
        else:
            raise CoordIOError('Invalid scale.')

        self._iers = None

        # Store the current time as a UNIX time. We'll use this later to know
        # how much time has passed and do a quick update to now.
        self.__internal_time = time.time() if value is None else None

    def __repr__(self):
        return f'<Time (JD={self.jd})>'

    @property
    def iers(self):
        """Gets the IERS table."""

        if self._iers is None:
            self._iers = IERS()

        return self._iers

    @property
    def jd1(self):
        """Returns the integer part of the Julian data."""

        return int(self.jd)

    @property
    def jd2(self):
        """Returns the fractional part of the Julian data."""

        return self.jd - self.jd1

    @property
    def mjd(self):
        """Returns the modified Julian date ``(JD-2400000.5)``."""

        return self.jd - 2400000.5

    def to_now(self, scale='UTC'):
        """Updates the internal value to the current TAI date.

        Parameters
        ----------
        scale : str
            The scale of the system clock. Assumes it's UTC but it can be
            changed if the internal clock is synced to TAI or other timescale.

        Returns
        -------
        now : `float`
            The new TAI JD. Also updates the internal value.

        """

        # If __internal_time is set this means we initialised this
        # time as now(). We can simply add how long it has passed
        # since then.
        if self.__internal_time:
            new_now = time.time()
            delta_jd = (new_now - self.__internal_time) / 86400.
            self.jd += delta_jd
            self.__internal_time = new_now
            return self.jd

        # Otherwise reinitialise the object.
        self.__init__(scale=scale)
        return self.jd

    def get_dut1(self):
        """Returns the delta UT1-UTC."""

        if not self.iers:
            raise CoordIOError('IERS table not loaded.')

        return self.iers.get_delta_ut1_utc(jd=self.jd)

    def to_utc(self):
        """Returns the date converted to JD in the UTC scale."""

        utc1 = ctypes.c_double()
        utc2 = ctypes.c_double()

        res = sofa.iauTaiutc(self.jd1, self.jd2, utc1, utc2)

        if res != 0:
            raise ValueError(f'iauTaiutc return with error code {res}.')

        return utc1.value + utc2.value

    def to_ut1(self):
        """Returns the date converted to JD in the UT1 scale."""

        # DTA is UT1-UTC, which can be expressed as delta_UT1-delta_AT.

        # delta_UT1=UT1-UTC and is given by the IERS tables, in seconds.
        # This will update the table if it's out of date.
        delta_ut1 = self.iers.get_delta_ut1_utc(self.jd)

        # delta_AT=TAI-UTC and can be obtained with the iauDat function but
        # it's easier to calculate it using iauTaiutc and subtracting TAI.
        delta_at = self.jd - self.to_utc()

        # delta_UT1-TAI, in seconds, for iauTaiut1
        dta = delta_ut1 - delta_at * 86400.

        ut1 = ctypes.c_double()
        ut2 = ctypes.c_double()

        res = sofa.iauTaiut1(self.jd1, self.jd2, dta, ut1, ut2)

        if res != 0:
            raise ValueError(f'iauTaiut1 return with error code {res}.')

        return ut1.value + ut2.value

    def to_tt(self):
        """Returns the date converted to JD in the TT scale."""

        return self.jd + 32.184 / 86400.

    def to_tdb(self, longitude=0.0, latitude=0.0, altitude=0.0):
        """Returns the date converted to JD in the TDB scale.

        Parameters
        ----------
        longitude : float
            The East-positive longitude of the site, in degrees.
        latitude : float
            The site latitude in degrees.
        altitude : float
            The altitude (elevation) of the site above see level, in meters.
            Defaults to zero meters.

        """

        # We need to call iauDtdb that models the delta between TT and TDB.
        # It depends on the location of the observer on the Earth, and the
        # fractional part of the UT1 time.

        # Earth radius in km
        RE = 6378.1

        rlong = numpy.radians(longitude)
        rlat = numpy.radians(latitude)

        # Distance from Earth spin axis (km)
        u = (RE + altitude / 1000.) * numpy.cos(rlat)

        # Distance north of equatorial plane (km)
        v = (RE + altitude / 1000.) * numpy.sin(rlat)

        ut1 = self.to_ut1()
        ut1_1 = ut1 - int(ut1)

        tt = self.to_tt()
        tt_1 = int(tt)
        tt_2 = tt - tt_1

        delta_tdb_tt = sofa.iauDtdb(tt_1, tt_2, ut1_1, rlong, u, v)

        return tt + delta_tdb_tt / 86400.



from __future__ import annotations

import numpy

from . import defaults
from .coordinate import Coordinate, Coordinate2D
from .exceptions import CoordIOError


class Guide(Coordinate2D):
    """Guide coordinates are 2D cartesian xy coordinates.  Units are pixels.
    The origin is the lower left corner of the lower left pixel when looking
    at the CCD through the camera window.  Thus guide images are taken in
    "selfie-mode".  So the coordinate (0.5, 0.5) is the center of the lower
    left pixel. -y points (generally) toward the wok vertex.  True rotations
    and locations of GFAs in the wok are measured/calibrated after
    installation and on-sky characterization.

    Parameters
    -----------
    value : numpy.ndarray
        A Nx2 array where [x,y] are columns.  Or `.Tangent`
    xBin : int
        CCD x bin factor, defaults to 1
    yBin : int
        CCD y bin factor, defaults to 1

    Attributes
    -----------
    guide_warn : numpy.ndarray
        boolean array indicating coordinates outside the CCD FOV
    """

    __extra_params__ = ["xBin", "yBin"]
    __warn_arrays__ = ["guide_warn"]

    xBin: int
    yBin: int
    guide_warn: numpy.ndarray

    def __new__(cls, value, **kwargs):

        xBin = kwargs.get("xBin", None)
        if xBin is None:
            kwargs["xBin"] = 1

        yBin = kwargs.get("yBin", None)
        if yBin is None:
            kwargs["yBin"] = 1

        if isinstance(value, Coordinate):
            if value.coordSysName == "Tangent":
                # going from 3D -> 2D coord sys
                # reduce dimensionality
                arrInit = numpy.zeros((len(value), 2))
                obj = super().__new__(cls, arrInit, **kwargs)
                obj._fromTangent(value)
            else:
                raise CoordIOError(
                    'Cannot convert to Guide from %s'%value.coordSysName
                )
        else:
            obj = super().__new__(cls, value, **kwargs)
            obj._fromRaw()

        return obj

    def _fromTangent(self, tangentCoords):
        """Convert from tangent coordinates to guide coordinates

        """
        if tangentCoords.holeID not in defaults.VALID_GUIDE_IDS:
            raise CoordIOError(
                "Cannot convert from wok hole %s to Guide" %
                tangentCoords.holeID
            )
        # make sure the guide wavelength is specified for all coords
        # this may be a little too extreme of a check
        badWl = numpy.sum(
            tangentCoords.wavelength - defaults.INST_TO_WAVE["GFA"]
        ) != 0
        if badWl:
            raise CoordIOError(
                "Cannont convert to Guide coords from non-guide wavelength"
            )

        # use coords projected to tangent xy plane
        # could almost equivalently use tangentCoords[:,0:2]
        # and we may want to ditch the projection eventually if its
        # unnecessary
        xT = tangentCoords.xProj
        yT = tangentCoords.yProj

        xPix = (1 / self.xBin) * (
            defaults.MICRONS_PER_MM / defaults.GFA_PIXEL_SIZE * xT + \
            defaults.GFA_CHIP_CENTER
        )

        yPix = (1 / self.yBin) * (
            defaults.MICRONS_PER_MM / defaults.GFA_PIXEL_SIZE * yT + \
            defaults.GFA_CHIP_CENTER
        )

        self[:, 0] = xPix
        self[:, 1] = yPix

        self._fromRaw()

    def _fromRaw(self):
        """Find coords that may be out of the FOV and tag them

        """

        # note this might be wrong when xBin is 3
        maxX = defaults.GFA_CHIP_CENTER * 2 / self.xBin
        maxY = defaults.GFA_CHIP_CENTER * 2 / self.yBin
        xPix = self[:, 0]
        yPix = self[:, 1]

        arg = numpy.argwhere((xPix < 0) | (xPix > maxX) | \
                             (yPix < 0) | (yPix > maxY))

        if len(arg) == 0:
            # everything in range...
            return

        arg = arg.squeeze()
        self.guide_warn[arg] = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-08-16
# @Filename: iers.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import os
import urllib.request
import warnings

import numpy

from . import config, sofa
from .exceptions import CoordIOError, CoordIOUserWarning


BASE_URL = 'ftp://ftp.iers.org/products/eop/rapid/'


class IERS:
    """Wrapper around the IERS bulletins.

    Parameters
    ----------
    path : str
        The path where the IERS table, in CSV format, lives or will be saved
        to. If `None`, defaults to ``config['iers']['path']``.
    channel : str
        The IERS channel to use. Right now only ``finals`` is implemented.
    download : bool
        If the IERS table is not found on disk or it's out of date, download
        an updated copy.

    Attributes
    ----------
    data : ~numpy.ndarray
        A record array with the parsed IERS data, trimmed to remove rows for
        which ``UT1-UTC`` is not defined.

    """

    __instance = None

    def __new__(cls, path=None, channel='finals', download=True):

        if cls.__instance is not None:

            if cls.__instance.is_valid():
                return cls.__instance
            else:
                # We take the default route which takes care of updating the
                # file.
                pass

        obj = super().__new__(cls)
        obj.data = None
        obj.channel = channel or config['iers']['channel']

        path = path or config['iers']['path']

        if channel == 'finals':
            obj.path = os.path.join(os.path.expanduser(path), 'finals2000A.data.csv')
        else:
            raise NotImplementedError('Only finals channels is implemented.')

        if os.path.exists(obj.path):
            cls.load_data(obj)
            if not cls.is_valid(obj):
                warnings.warn('Current IERS file is out of date. '
                              'Redownloading.', CoordIOUserWarning)
                cls.update_data(obj, channel=obj.channel)
        else:
            cls.update_data(obj, channel=obj.channel)

        # Check data one last time.
        if not cls.is_valid(obj):
            raise CoordIOError('IERS table is not valid. '
                               'This should not have happened.')

        cls.__instance = obj

        return cls.__instance

    def update_data(self, path=None, channel=None):
        """Update the IERS table, downloading the latest available version."""

        path = path or self.path
        channel = channel or self.channel

        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))

        if channel == 'finals':
            URL = BASE_URL + 'standard/csv/finals.data.csv'
            warnings.warn(f'Downloading IERS table from {URL}.', CoordIOUserWarning)
            url = urllib.request.urlopen(URL)
            data = url.read()

            with open(path, 'wb') as fd:
                fd.write(data)
            url.close()
        else:
            raise NotImplementedError('Only finals channels is implemented.')

        self.load_data(path=path)

    def _get_current_jd(self):
        """Returns the current JD in the scale of the computer clock."""

        jd1, jd2 = sofa.get_internal_date()
        jd = jd1 + jd2

        return jd

    def is_valid(self, jd=None, offset=5, download=True):
        """Determines whether a JD is included in the loaded IERS table.

        Parameters
        ----------
        jd : float
            The Julian date to check. There is a certain ambiguity in the
            format of the MJD in the IERS tables but for most purposes the
            difference between UTC and TAI should be meaningless for these
            purposes.
        offset : int
            Number of days to look ahead. If ``jd+offset`` is not included in
            the IERS table the method will return `False`.
        download : bool
            Whether to automatically download an updated version of the IERS
            table if the requested date is not included in the current one.

        Returns
        -------
        is_valid : `bool`
            Whether the date is valid within the current (or newly downloaded)
            IERS table.

        """

        if self.data is None:
            raise CoordIOError('IERS data has not been loaded')

        if jd is None:
            jd = self._get_current_jd()

        mjd = int(jd - 2400000.5)

        min_mjd = self.data['MJD'].min()
        max_mjd = self.data['MJD'].max()

        if mjd - offset > min_mjd and mjd + offset < max_mjd:
            return True
        else:
            if download:
                self.update_data()
                return self.is_valid(jd, offset, download=False)
            return False

    def load_data(self, path=None):
        """Loads an IERS table in CSV format from ``path``."""

        path = path or self.path

        if not os.path.exists(path):
            raise FileNotFoundError(f'File {path!r} not found.')

        self.data = numpy.genfromtxt(path, delimiter=';', names=True,
                                     dtype=None, encoding='UTF-8')

        # Trim rows without UT1-UTC data
        self.data = self.data[~numpy.isnan(self.data['UT1UTC'])]

    def get_delta_ut1_utc(self, jd=None, download=True):
        """Returns the interpolated ``UT1-UTC`` value, in seconds."""

        if jd is None:
            jd = self._get_current_jd()

        if not self.is_valid(jd, offset=0, download=download):
            raise CoordIOError('IERS table is out of date.')

        # Whats the timescale for MJD in the IERS tables?
        # Probably doesn't matter.
        mjd = jd - 2400000.5

        return numpy.interp(mjd, self.data['MJD'], self.data['UT1UTC'])



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy

from . import conv, defaults
from .conv import cart2FieldAngle, sph2Cart
from .coordinate import (Coordinate, Coordinate2D, Coordinate3D,
                         verifySite, verifyWavelength)
from .exceptions import CoordIOError


if TYPE_CHECKING:
    from .site import Site
    from .sky import Observed


__all__ = ["Field", "FocalPlane"]


class Field(Coordinate2D):
    """A representation of Field Coordinates.  A spherical coordinate system
    defined by two angles: theta, phi.  Theta is the angle about the optical
    axis measured from the direction of +RA. Phi is the angle off the optical
    axis of the telescope.  So Phi=0 is the optical axis of the telescope
    increasing from earth to sky.

    In the Cartesian representation on a unit sphere, +x is aligned with +RA,
    +y is aligned with +Dec, and so +z is aligned with the telescope's optical
    axis and points from the earth to sky.

    Parameters
    ------------
    value : numpy.ndarray
        A Nx2 array.  First column is theta, the azimuthal angle from +RA
        through +Dec in deg. Second angle is phi, polar angle angle (off axis
        angle) in deg. Or `.Observed`.  Or `.FocalPlane`.

    field_center : `.Observed`
        An `.Observed` instance containing a single coordinate

    Attributes
    ------------
    x : numpy.ndarray
        unit-spherical x coordinate
    y : numpy.ndarray
        unit-spherical y coordinate
    z : numpy.ndarray
        unit-spherical z coordinate
    x_angle : numpy.ndarray
        zeemax-style x field angle (deg)
    y_angle : numpy.ndarray
        zeemax-style y field angle (deg)
    field_warn: numpy.ndarray
        boolean array indicating suspect conversions from large field angles
    """

    __computed_arrays__ = ["x", "y", "z", "x_angle", "y_angle"]
    __warn_arrays__ = ["field_warn"]
    __extra_params__ = ["field_center"]  # mandatory parameter
    # may want to carry around position angle for all
    # coordinates too through the chain?  Could reduce errors in guiding
    # because direction to north or zenith varies across the field due to...
    # spheres.  For now ignore it?

    x: numpy.ndarray
    y: numpy.ndarray
    z: numpy.ndarray
    x_angle: numpy.ndarray
    y_angle: numpy.ndarray
    field_warn: numpy.ndarray
    field_center: Observed

    def __new__(cls, value, **kwargs):

        field_center = kwargs.get("field_center", None)
        if field_center is None:
            raise CoordIOError("field_center must be passed to Field")
        else:
            if not hasattr(field_center, "coordSysName"):
                raise CoordIOError(
                    "field_center must be an Observed coordinate"
                )
            if field_center.coordSysName != "Observed":
                raise CoordIOError(
                    "field_center must be an Observed coordinate"
                )
            if len(field_center) != 1:
                raise CoordIOError("field_center must contain only one coord")

        if isinstance(value, Coordinate):
            if value.coordSysName == "Observed":
                obj = super().__new__(cls, value, **kwargs)
                obj._fromObserved(value)
            elif value.coordSysName == "FocalPlane":
                # focal plane is a 3D coord sys
                # field is 2D (spherical)
                initArray = numpy.zeros((len(value), 2))
                obj = super().__new__(cls, initArray, **kwargs)
                obj._fromFocalPlane(value)
            else:
                raise CoordIOError(
                    "Cannot convert to Field from %s" % value.coordSysName
                )
        else:
            obj = super().__new__(cls, value, **kwargs)
            obj._fromRaw()

        return obj

    def _fromObserved(self, obsCoords):
        """Converts from observed coords to field coords, given the field
        center.  Populates the computed arrays

        Parameters
        -----------
        obsCoords : `.Observed`
        """
        altAz = numpy.array(obsCoords)
        alt, az = altAz[:, 0], altAz[:, 1]
        altCenter, azCenter = self.field_center.flatten()
        pa = float(self.field_center.pa)  # position angle
        theta, phi = conv.observedToField(alt, az, altCenter, azCenter, pa)
        self[:, 0] = theta
        self[:, 1] = phi

        self._fromRaw()

    def _fromFocalPlane(self, fpCoords):
        """Convert from FocalPlane coords to Field coords.

        Parameters
        ------------
        fpCoords : `.FocalPlane`
        """
        siteName = fpCoords.site.name.upper()
        for waveCat in ["Boss", "Apogee", "GFA"]:
            # find which coords are intended for which
            # wavelength
            arg = numpy.argwhere(
                fpCoords.wavelength == defaults.INST_TO_WAVE[waveCat]
            )

            if len(arg) == 0:
                continue

            arg = arg.squeeze()

            xFP = fpCoords[arg, 0].squeeze()
            yFP = fpCoords[arg, 1].squeeze()
            zFP = fpCoords[arg, 2].squeeze()

            if hasattr(xFP, "__len__") and len(xFP) == 0:
                continue

            thetaFocal, phiFocal, fieldWarn = conv.focalToField(
                xFP,
                yFP,
                zFP,
                siteName,
                waveCat
            )
            thetaPhi = numpy.array([thetaFocal, phiFocal]).T
            self[arg] = thetaPhi
            self.field_warn[arg] = fieldWarn

        self._fromRaw()

    def _fromRaw(self):
        """Populates the computed arrays
        """
        # ensure wrapping
        self[:,0] = self[:,0] % 360
        self.x, self.y, self.z = sph2Cart(self[:, 0], self[:, 1])
        self.x_angle, self.y_angle = cart2FieldAngle(self.x, self.y, self.z)


class FocalPlane(Coordinate3D):
    """The focal plane coordinate system is 3D Cartesian with units of mm.
    The origin is the M1 vertex.  +x is aligned with +RA on the image, +y is
    aligned with +Dec on the image, +z points from earth to sky along the
    telescope's optical axis. `.FocalPlane` xy's are physically rotated 180
    degrees with respect to `.Field` xy's because Cassegrain telescope rotates
    the image.  The choice was made to put +x along +RA on the image, rather
    than the sky.

    For each wavelength, a sphereical surface fit has been determined.
    The sphere's center lies along the optical axis at a positive z value.
    The sphere's center is considered the 'ray origin', so all rays traveling
    to the focal plane originate from this location.

    The ray origin and a wavelength-dependent, z-symmetric, distortion model
    is used to convert between `.FocalPlane` and `.Field` coordinates.
    The distortion model is an odd-termed polynomial relating input field
    angle to output focal plane model.

    The ZEMAX models/scripts, and fitting code for this still lives in
    github.com/csayres/sdssconv, but may be eventually moved to this package.

    Parameters
    ------------
    value : numpy.ndarray
        A Nx3 array where [x,y,z] are columns. Or `.Field`.  Or `.Wok`.
    wavelength : float or numpy.ndarray
        A 1D array with he observed wavelength, in angstroms.
        Currently only values of 5400, 6231, and 16600 are valid, corresponding
        to BOSS, Apogee, and sdss-r band (GFA).  Defaults to GFA wavelength.
    site : `.Site`
        Used to pick the focal plane model to use which is telescope dependent

    Attributes
    -----------
    b : numpy.ndarray
        A 1D array containing the location of the spherical fit's center
        (origin of rays) along the optical axis in mm.  Varies with wavelength.
    R : numpy.ndarray
        1D array containing the radius of curvature for the spherical fit in
        mm.  Varies with wavelength.
    field_warn: numpy.ndarray
        boolean array indicating suspect conversions from large field angles
    """

    __extra_params__ = ["site"]   # mandatory argument
    __extra_arrays__ = ["wavelength"]
    __computed_arrays__ = ["b", "R"]
    __warn_arrays__ = ["field_warn"]

    b: numpy.ndarray
    R: numpy.ndarray
    field_warn: numpy.ndarray
    wavelength: numpy.ndarray
    site: Site

    def __new__(cls, value, **kwargs):

        verifySite(kwargs)

        verifyWavelength(
            kwargs, len(value), strict=True
        )

        if isinstance(value, Coordinate):
            if value.coordSysName == "Field":
                # going from 2D -> 3D coord sys
                # expand dimensionality
                arrInit = numpy.zeros((len(value), 3))
                obj = super().__new__(cls, arrInit, **kwargs)
                obj._fromField(value)
            elif value.coordSysName == "Wok":
                # wok is already a 3D coord sys
                obj = super().__new__(cls, value, **kwargs)
                obj._fromWok(value)
            else:
                raise CoordIOError(
                    'Cannot convert to FocalPlane from %s'%value.coordSysName
                )
        else:
            obj = super().__new__(cls, value, **kwargs)
            obj._fromRaw()

        return obj

    def _fromField(self, fieldCoord):
        """ Convert from field coordinates to focal plane coordinates, using
        a pre-fit focal plane model.

        Parameters
        -----------
        fieldCoord : `.Field`
        """
        siteName = self.site.name.upper()
        for waveCat in ["Boss", "Apogee", "GFA"]:
            # find which coords are intended for which
            # wavelength
            arg = numpy.argwhere(
                self.wavelength == defaults.INST_TO_WAVE[waveCat]
            )

            if len(arg) == 0:
                continue

            arg = arg.squeeze()

            thetaField = fieldCoord[arg, 0].squeeze()
            phiField = fieldCoord[arg, 1].squeeze()

            xFP, yFP, zFP, R, b, fieldWarn = conv.fieldToFocal(
                thetaField,
                phiField,
                siteName,
                waveCat
            )

            xyzFP = numpy.array([xFP, yFP, zFP]).T
            self[arg] = xyzFP
            self.b[arg] = b
            self.R[arg] = R
            self.field_warn[arg] = fieldWarn

    def _fromWok(self, wokCoord):
        """ Convert from wok coordinates to focal plane coordinates.

        Parameters
        -----------
        wokCoord : `.Wok`
        """
        pa = wokCoord.obsAngle
        siteName = self.site.name.upper()
        xOff, yOff, zOff, tiltX, tiltY = defaults.getWokOrient(siteName)
        xWok, yWok, zWok = wokCoord[:, 0], wokCoord[:, 1], wokCoord[:, 2]
        xF, yF, zF = conv.wokToFocal(
            xWok, yWok, zWok, pa, xOff, yOff, zOff, tiltX, tiltY
        )

        self[:, 0] = xF
        self[:, 1] = yF
        self[:, 2] = zF

        self._fromRaw()

    def _fromRaw(self):
        direction = "focal"  # irrelevant, just grabbing sphere params

        for waveCat in ["Boss", "Apogee", "GFA"]:
            arg = numpy.argwhere(
                self.wavelength == defaults.INST_TO_WAVE[waveCat])

            if len(arg) == 0:
                continue

            arg = arg.squeeze()

            R, b, c0, c1, c2, c3, c4 = defaults.getFPModelParams(direction, waveCat)
            self.b[arg] = b
            self.R[arg] = R



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-08-16
# @Filename: sofa_bindings.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

# Bindings for SOFA procedures. This does not intent to be a comprehensive
# wrapping of the library, just of those functions that are needed for our
# purposes.

import ctypes
import datetime
import importlib
import os
from ctypes import POINTER, c_char_p, c_double, c_int


# List of argument times for some of the SOFA functions.
ARGTYPES = [

    # const char *scale, int iy, int im, int id,
    #   int ihr, int imn, double sec,
    #   double *d1, double *d2
    ('iauDtf2d', (c_char_p, c_int, c_int, c_int,
                  c_int, c_int, c_double,
                  POINTER(c_double), POINTER(c_double))),

    # double utc1, double utc2, double *tai1, double *tai2
    ('iauUtctai', (c_double, c_double, POINTER(c_double), POINTER(c_double))),

    # double tai1, double tai2, double dta, double *ut11, double *ut12
    ('iauTaiut1', (c_double, c_double, c_double,
                   POINTER(c_double), POINTER(c_double))),

    # double tai1, double tai2, double *utc1, double *utc2
    ('iauTaiutc', (c_double, c_double, POINTER(c_double), POINTER(c_double))),

    # int iy, int im, int id, double fd, double *deltat
    ('iauDat', (c_int, c_int, c_int, c_double, POINTER(c_double))),

    # double rc, double dc, double pr, double pd, double px, double rv,
    # double utc1, double utc2, double dut1,
    # double elong, double phi, double hm, double xp, double yp,
    # double phpa, double tc, double rh, double wl,
    # double *aob, double *zob, double *hob,
    # double *dob, double *rob, double *eo
    ('iauAtco13', (c_double, c_double, c_double, c_double, c_double, c_double,
                   c_double, c_double, c_double,
                   c_double, c_double, c_double, c_double, c_double,
                   c_double, c_double, c_double, c_double,
                   POINTER(c_double), POINTER(c_double), POINTER(c_double),
                   POINTER(c_double), POINTER(c_double), POINTER(c_double))),


    ('iauAtoc13', (c_char_p, c_double, c_double, c_double, c_double, c_double,
                   c_double, c_double, c_double, c_double, c_double, c_double,
                   c_double, c_double, c_double, POINTER(c_double), POINTER(c_double))),

    # double date1, double date2, double ut, double elong, double u, double v
    ('iauDtdb', (c_double, c_double, c_double, c_double, c_double, c_double)),

    # double ra1, double dec1, double pmr1, double pmd1,
    # double px1, double rv1,
    # double ep1a, double ep1b, double ep2a, double ep2b,
    # double *ra2, double *dec2, double *pmr2, double *pmd2,
    # double *px2, double *rv2
    ('iauPmsafe', (c_double, c_double, c_double, c_double, c_double, c_double,
                   c_double, c_double, c_double, c_double,
                   POINTER(c_double), POINTER(c_double), POINTER(c_double),
                   POINTER(c_double), POINTER(c_double), POINTER(c_double))),

    # compute position angle http://www.iausofa.org/2019_0722_C/sofa/hd2pa.c
    # args: hour angle, declination, site latitude (radians)
    ('iauHd2pa', (c_double, c_double, c_double)),

    # Horizon to equatorial coordinates: transform azimuth and altitude
    # to hour angle and declination.
    # args: double az, double alt, double latitude, double *ha, double *dec
    ('iauAe2hd', (c_double, c_double, c_double,
                  POINTER(c_double), POINTER(c_double))),

    # Earth rotation angle
    # sum of the arguments should be jd in the UT1 scale
    # args: double jd, double 0
    ('iauEra00', (c_double, c_double))

]


class SOFA(ctypes.CDLL):
    """Load the SOFA C extension.

    Parameters
    ----------
    name : str
        The path to the shared object. If `None`, will try the standard
        variants of ``libsofa`` for this architecture. If the library is not
        in a standard location the path needs to be added to
        ``LD_LIBRARY_PATHS``.

    """

    def __init__(self, name=None):

        if name is None:
            # Try all the usual suffixes.
            success = False
            mod_path = os.path.join(os.path.dirname(__file__), 'libsofa')
            for suffix in importlib.machinery.EXTENSION_SUFFIXES:
                try:
                    super().__init__(mod_path + suffix)
                    success = True
                    break
                except OSError:
                    pass
            if success is False:
                raise OSError('Could not find a valid libsofa extension.')
        else:
            super().__init__(name)

        # Assign argtypes. This also adds the functions to the class dir().
        for func_name, argtypes in ARGTYPES:
            func = self.__getattr__(func_name)
            func.argtypes = argtypes

        # Some special cases of function that do not return
        # an integer error check.
        self.iauDtdb.restype = c_double
        self.iauHd2pa.restype = c_double
        self.iauEra00.restype = c_double

    def get_internal_date(self, date=None, scale='UTC'):
        """Returns the internal representation of a date.

        Parameters
        ----------
        date : ~datetime.datetime
            The date to convert. If `None`, `~datetime.datetime.utcnow` will
            be used and ``scale`` will be ignored.
        scale : str
            The scale of the date.

        Returns
        -------
        d1, d2 : `float`
            The two-part Julian date.

        """

        if date is None:
            date = datetime.datetime.utcnow()
            scale = 'UTC'

        d1 = c_double()
        d2 = c_double()

        res = self.iauDtf2d(scale.upper().encode(), date.year, date.month,
                            date.day, date.hour, date.minute,
                            date.second + date.microsecond / 1e6,
                            d1, d2)

        if res != 0:
            raise ValueError(f'iauDtf2d return with error code {res}.')

        return d1.value, d2.value



# encoding: utf-8
# isort:skip_file

import os
import warnings

from sdsstools import get_config, get_logger, get_package_version

from .exceptions import CoordIOError


# PYPI package name
NAME = 'sdss-coordio'

# Loads config. For coordio we don't allow overridding with user config files.
config = get_config('coordio', allow_user=False)

# Inits the logging system as NAME.
log = get_logger(NAME)

# Package name should be PYPI package name
__version__ = get_package_version(path=__file__, package_name=NAME)


from .exceptions import CoordIOUserWarning
from .sofa_bindings import SOFA

try:
    sofa = SOFA()
except OSError:
    warnings.warn('Cannot load libsofa. '
                  'Most functionality will not be available.',
                  CoordIOUserWarning)
    sofa = None


from .exceptions import *
from .iers import IERS
from .site import Site
from .sky import *
from .telescope import *
from .time import Time
from .wok import Wok
from .tangent import Tangent
from .guide import Guide
from .positioner import PositionerBoss, PositionerApogee, PositionerMetrology
from .utils import *
from .fitData import TransRotScaleModel



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-08-14
# @Filename: build.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

# isort:skip_file

# Order matters here. Keep this import at the top.
from setuptools import setup

import glob
import sys
from distutils.core import Extension


LIBSOFA_PATH = 'cextern/sofa'
LIBCOORDIO_PATH = 'cextern/conv.cpp'


extra_compile_args = ['-c', '-pedantic', '-Wall', '-W', '-O']
extra_link_args = []


class getPybindInclude(object):
    """Helper class to determine the pybind11 include path
    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked.
    https://github.com/pybind/python_example/blob/master/setup.py
    """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)


sofa_sources = glob.glob(LIBSOFA_PATH + '/*.c')
includes = [
    'include',
    'src/coordio/include',
    getPybindInclude(),
    getPybindInclude(user=True)
]


extra_compile_args2 = ["--std=c++11", "-fPIC", "-v", "-O3"]
extra_link_args2 = None

if sys.platform == 'darwin':
    extra_compile_args2 += ['-stdlib=libc++', '-mmacosx-version-min=10.9']
    extra_link_args2 = ["-v", '-mmacosx-version-min=10.9']

    from distutils import sysconfig
    vars = sysconfig.get_config_vars()
    vars['LDSHARED'] = vars['LDSHARED'].replace('-bundle', '-dynamiclib')


ext_modules = [
    Extension(
        'coordio.libsofa',
        sources=sofa_sources,
        include_dirs=[LIBSOFA_PATH],
        libraries=[],
        define_macros=[],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        language='c',
        optional=False),
    Extension(
        'coordio.libcoordio',
        sources=[LIBCOORDIO_PATH],
        include_dirs=includes,
        extra_compile_args=extra_compile_args2,
        extra_link_args=extra_link_args2,
        optional=False),
]

setup(ext_modules=ext_modules)



import numpy

from .sky import ICRS, Observed
from .telescope import Field, FocalPlane
from .wok import Wok
from .site import Site
from . import defaults


def wokCurveAPO(r):
    """Curve of the wok at APO at radial position r

    Parameters
    -----------
    r : scalar or 1D array
        radius (cylindrical coords) mm

    Returns:
    ---------
    result : scalar or 1D array
        z (height) of wok surface in mm (0 at vertex)
    """
    A = 9199.322517101522
    return A - numpy.sqrt(A**2 - r**2)


def wokCurveLCO(r):
    """Curve of the wok LCO at radial position r

    Parameters
    -----------
    r : scalar or 1D array
        radius (cylindrical coords) mm

    Returns:
    ---------
    result : scalar or 1D array
        z (height) of wok surface in mm (0 at vertex)
    """
    A = 0.000113636363636
    B = 0.0000000129132231405
    C = 0.0000012336318
    return A * r**2 / (1 + numpy.sqrt(1 - B * r**2)) + C * r**2


def radec2wokxy(ra, dec, coordEpoch, waveName, raCen, decCen, obsAngle,
                obsSite, obsTime, pmra=None, pmdec=None, parallax=None,
                radVel=None, pressure=None, relativeHumidity=0.5,
                temperature=10):
    r"""
    Convert from ra, dec ICRS coords to a flat-wok XY in mm.  At obsAngle=0
    wok +y is a aligned with +dec, and wok +x is aligned with +ra

    Question for José, do we need to enforce a time scale?  I think everything
    defaults to UTC.

    Parameters
    ------------
    ra : numpy.ndarray
        Right ascension in ICRS, degrees
    dec : numpy.ndarray
        Declination in ICRS, degrees
    coordEpoch : float
        A TDB Julian date, the epoch for the input
        coordinates (from the catalog). Defaults to J2000.
    waveName : str, or numpy.ndarray
        Array elements must be "Apogee", "Boss", or "GFA" strings
    raCen : float
        Right ascension of field center, in degrees
    decCen : float
        Declination of field center, in degrees.
    obsAngle : float
        Position angle for observation.  Angle is measured from North
        through East to wok +y. So obsAngle of 45 deg, wok +y points NE.
    obsSite : str
        Either "APO" or "LCO"
    obsTime : float
        TDB Julian date.  The time at which these coordinates will be observed
        with the FPS.
    pmra : numpy.ndarray
        A 1D array with the proper motion in the RA axis for the N targets,
        in milliarcsec/yr. Must be a true angle, i.e, it must include the
        ``cos(dec)`` term.  Defaults to 0.
    pmdec : numpy.ndarray
        A 1D array with the proper motion in the RA axis for the N targets,
        in milliarcsec/yr.  Defaults to 0.
    parallax : numpy.ndarray
        A 1D array with the parallax for the N targets, in milliarcsec.
        Defaults to 0.
    radVel : numpy.ndarray
        A 1D array with the radial velocity in km/s, positive when receding.
        Defaults to 0.
    pressure : float
        The atmospheric pressure at the site, in millibar (same as hPa). If
        not provided the pressure will be calculated using the altitude
        :math:`h` and the approximate expression

        .. math::

            p \sim -p_0 \exp\left( \dfrac{g h M} {T_0 R_0} \right)

        where :math:`p_0` is the pressure at sea level, :math:`M` is the molar
        mass of the air, :math:`R_0` is the universal gas constant, and
        :math:`T_0=288.16\,{\rm K}` is the standard sea-level temperature.
    relativeHumidity : float
        The relative humidity, in the range :math:`0-1`. Defaults to 0.5.
    temperature : float
        The site temperature, in degrees Celsius. Defaults to
        :math:`10^\circ{\rm C}`.

    Returns
    ---------
    xWok : numpy.ndarray
        x wok coordinate, mm
    yWok : numpy.ndarray
        y wok coordinate, mm
    fieldWarn : numpy.ndarray
        boolean array.  Where True the coordinate converted should be eyed with
        suspicion.  (It came from very far off axis).
    hourAngle : float
        hour angle of field center in degrees
    positionAngle : float
        position angle of field center in degrees
    """
    nCoords = len(ra)

    # first grab the correct wavelengths for fibers
    wavelength = numpy.zeros(nCoords)

    if isinstance(waveName, str):
        # same wl for all coords
        wavelength += defaults.INST_TO_WAVE[waveName]
    else:
        assert len(waveName) == nCoords
        for ii, ft in enumerate(waveName):
            wavelength[ii] = defaults.INST_TO_WAVE[ft]

    site = Site(
        obsSite, pressure=pressure,
        temperature=temperature, rh=relativeHumidity
    )
    site.set_time(obsTime)

    # first determine the field center in observed coordinates
    # use the guide wavelength for field center
    # epoch not needed, no propermotions, etc (josé verify?)
    icrsCen = ICRS([[raCen, decCen]])
    obsCen = Observed(icrsCen, site=site, wavelength=defaults.INST_TO_WAVE["GFA"])


    radec = numpy.array([ra, dec]).T

    icrs = ICRS(
        radec, epoch=coordEpoch, pmra=pmra, pmdec=pmdec,
        parallax=parallax, rvel=radVel
    )

    # propogate propermotions, etc
    icrs = icrs.to_epoch(obsTime, site=site)

    obs = Observed(icrs, site=site, wavelength=wavelength)
    field = Field(obs, field_center=obsCen)
    focal = FocalPlane(field, wavelength=wavelength, site=site)
    wok = Wok(focal, site=site, obsAngle=obsAngle)

    output = (
        wok[:, 0], wok[:, 1], focal.field_warn,
        float(obsCen.ha), float(obsCen.pa)
    )
    return output


def wokxy2radec(xWok, yWok, waveName, raCen, decCen, obsAngle,
                obsSite, obsTime, pressure=None, relativeHumidity=0.5,
                temperature=10):
    r"""
    Convert from flat-wok XY (mm) to ra, dec ICRS coords (deg)

    Question for José, do we need to enforce a time scale?  I think everything
    defaults to UTC.

    Parameters
    ------------
    xWok : numpy.ndarray
        X wok coordinates, mm
    yWok : numpy.ndarray
        Y wok coordinates, mm
    waveName : str, or numpy.ndarray
        Array elements must be "Apogee", "Boss", or "GFA" strings
    raCen : float
        Right ascension of field center, in degrees
    decCen : float
        Declination of field center, in degrees.
    obsAngle : float
        Position angle for observation.  Angle is measured from North
        through East to wok +y. So obsAngle of 45 deg, wok +y points NE.
    obsSite : str
        Either "APO" or "LCO"
    obsTime : float
        TDB Julian date.  The time at which these coordinates will be observed
        with the FPS.
    pressure : float
        The atmospheric pressure at the site, in millibar (same as hPa). If
        not provided the pressure will be calculated using the altitude
        :math:`h` and the approximate expression

        .. math::

            p \sim -p_0 \exp\left( \dfrac{g h M} {T_0 R_0} \right)

        where :math:`p_0` is the pressure at sea level, :math:`M` is the molar
        mass of the air, :math:`R_0` is the universal gas constant, and
        :math:`T_0=288.16\,{\rm K}` is the standard sea-level temperature.
    relativeHumidity : float
        The relative humidity, in the range :math:`0-1`. Defaults to 0.5.
    temperature : float
        The site temperature, in degrees Celsius. Defaults to
        :math:`10^\circ{\rm C}`.

    Returns
    ---------
    xWok : numpy.ndarray
        x wok coordinate, mm
    yWok : numpy.ndarray
        y wok coordinate, mm
    fieldWarn : numpy.ndarray
        boolean array.  Where True the coordinate converted should be eyed with
        suspicion.  (It came from very far off axis).
    """
    nCoords = len(xWok)

    # first grab the correct wavelengths for fibers
    wavelength = numpy.zeros(nCoords)

    if isinstance(waveName, str):
        # same wl for all coords
        wavelength += defaults.INST_TO_WAVE[waveName]
    else:
        assert len(waveName) == nCoords
        for ii, ft in enumerate(waveName):
            wavelength[ii] = defaults.INST_TO_WAVE[ft]

    site = Site(
        obsSite, pressure=pressure,
        temperature=temperature, rh=relativeHumidity
    )
    site.set_time(obsTime)

    # first determine the field center in observed coordinates
    # use the guide wavelength for field center
    # epoch not needed, no propermotions, etc (josé verify?)
    icrsCen = ICRS([[raCen, decCen]])
    obsCen = Observed(icrsCen, site=site, wavelength=defaults.INST_TO_WAVE["GFA"])

    # hack in z wok position of 143 mm
    # could do a little better and estimate the curve of focal surface
    xyzWok = numpy.zeros((nCoords, 3))
    rWok = numpy.sqrt(xWok**2 + yWok**2)

    # this is not totally correct
    # but probably better than modeling as flat
    # project the xy positions onto the 3d surface
    # and add 143 (the positioner height)
    if obsSite == "APO":
        zWok = defaults.POSITIONER_HEIGHT + wokCurveAPO(rWok)
    else:
        zWok = defaults.POSITIONER_HEIGHT + wokCurveLCO(rWok)

    xyzWok[:, 0] = xWok
    xyzWok[:, 1] = yWok
    xyzWok[:, 2] = zWok

    wok = Wok(xyzWok, site=site, obsAngle=obsAngle)
    focal = FocalPlane(wok, wavelength=wavelength, site=site)
    field = Field(focal, field_center=obsCen)
    obs = Observed(field, site=site, wavelength=wavelength)
    icrs = ICRS(obs, epoch=obsTime)

    return icrs[:, 0], icrs[:, 1], field.field_warn



from __future__ import annotations

from typing import TYPE_CHECKING

from . import conv, defaults
from .coordinate import Coordinate, Coordinate3D, verifySite
from .exceptions import CoordIOError


if TYPE_CHECKING:
    from .site import Site


class Wok(Coordinate3D):
    """A representation of a Wok Coordinates.  A 3D Cartesian coordinate
    system.  The orgin is the vertex of the Wok. -x points toward the Boss
    slithead.  +z points from earth to sky.  +y is dictated by right hand
    rule (so it points toward GFA-S2).

    A rotation about focal coords' +z is applied to observe fields at non-zero position
    angles

    With perfect alignment to the rotator at zero position angle, the xyz
    axes of the wok are idenentical to those of the focal plane system.
    The origin of the wok is translated along z with respect to the focal
    plane system by an amount determined by the ZEMAX models of the focal
    plane.

    True translations and tilts of the wok with respect to the focal plane
    are measured on sky (at zero position angle) and stored in the
    ``etc/wokOrientation.csv`` file.

    Parameters
    ------------
    value : numpy.ndarray
        A Nx3 array where [x,y,z] are columns. Or `.FocalPlane`.  Or `.Tangent`.
    site : `.Site`
        site name determines which wok parameters to use.  Mandatory parameter.
    obsAngle : float
        Position angle of observation. Angle measured from (image) North
        through East to wok +y. So obsAngle of 45 deg, wok +y points NE.
        Defaults to zero.

    """

    __extra_params__ = ["site", "obsAngle"]

    site: Site
    obsAngle: float

    def __new__(cls, value, **kwargs):

        verifySite(kwargs, strict=False)

        obsAngle = kwargs.get("obsAngle", None)
        if obsAngle is None:
            kwargs["obsAngle"] = 0.0

        obj = super().__new__(cls, value, **kwargs)

        if isinstance(value, Coordinate):
            if value.coordSysName == "FocalPlane":
                obj._fromFocalPlane(value)
            elif value.coordSysName == "Tangent":
                obj._fromTangent(value)
            else:
                raise CoordIOError(
                    "Cannot convert to Wok from %s" % value.coordSysName
                )

        return obj

    def _fromFocalPlane(self, fpCoords):
        """Converts from focal plane coords to wok coords

        Parameters:
        -------------
        fpCoords : `.FocalPlane`

        """
        xOff, yOff, zOff, tiltX, tiltY = defaults.getWokOrient(self.site.name)
        xWok, yWok, zWok = conv.focalToWok(
            fpCoords[:, 0], fpCoords[:, 1], fpCoords[:, 2],
            self.obsAngle, xOff, yOff, zOff, tiltX, tiltY
        )

        self[:, 0] = xWok
        self[:, 1] = yWok
        self[:, 2] = zWok

    def _fromTangent(self, tangentCoords):
        """Converts from tangent coords to wok coords

        Parameters:
        -------------
        tangentCoords : `.Tangent`

        """
        tx = tangentCoords[:, 0]
        ty = tangentCoords[:, 1]
        tz = tangentCoords[:, 2]

        b, iHat, jHat, kHat = defaults.getHoleOrient(
            tangentCoords.site.name, tangentCoords.holeID
        )

        xWok, yWok, zWok = conv.tangentToWok(
            tx, ty, tz, b, iHat, jHat, kHat, scaleFac=tangentCoords.scaleFactor
        )

        self[:, 0] = xWok
        self[:, 1] = yWok
        self[:, 2] = zWok
